// Vikunja is a to-do list application to facilitate your life.
// Copyright 2018-2020 Vikunja and contributors. All rights reserved.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package models

import (
	"time"
)

// OAuth2BlockedJTI contains all JTI signatures for already assigned JWT tokens
type OAuth2BlockedJTI struct {
	Signature string `xorm:"varchar(64) pk not null"`
	ExpiresAt *time.Time
}

// TableName returns the table's name
func (jti OAuth2BlockedJTI) TableName() string {
	return "oauth2_blocked_jtis"
}

// Expired checks if the JTI already expired
func (jti *OAuth2BlockedJTI) Expired() bool {
	return jti.ExpiresAt.Before(time.Now())
}

// BlockOAuth2JTI blocks the given JTI from reassigning until it expires
func BlockOAuth2JTI(jti string, expiresAt *time.Time) error {
	_, err := x.Insert(&OAuth2BlockedJTI{
		Signature: calculateSignature(jti),
		ExpiresAt: expiresAt,
	})
	return err
}

// DeleteExpiredJTIs deletes all expired JTIs from the database
func DeleteExpiredJTIs() error {
	_, err := x.Where("expires_at < now()").Delete(&OAuth2BlockedJTI{})
	return err
}

// IsJTIBlocked checks if a given JTI is already present in the block list and if present checks if the present
// JTI is expired. True is the default value if an error occurs.
func IsJTIBlocked(jti string) (bool, error) {
	blockedJTI := new(OAuth2BlockedJTI)
	if has, err := x.ID(calculateSignature(jti)).Get(&blockedJTI); err != nil {
		return true, err
	} else if !has {
		return false, nil
	}
	return blockedJTI.Expired(), nil
}
