// Vikunja is a to-do list application to facilitate your life.
// Copyright 2018-2020 Vikunja and contributors. All rights reserved.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package models

import "time"

// OAuth2SessionType represents the type of the session
type OAuth2SessionType string

const (
	// SessionTypeAuthorizationCode represents an authorization code session
	SessionTypeAuthorizationCode OAuth2SessionType = "code"
	// SessionTypeAccessToken represents an access token session
	SessionTypeAccessToken OAuth2SessionType = "access_token"
	// SessionTypeRefreshToken represents a refresh token session
	SessionTypeRefreshToken OAuth2SessionType = "refresh_token"
)

// OAuth2Session is a session used in the oauth2 process
type OAuth2Session struct {
	Signature string `xorm:"varchar(64) pk not null"`

	Type OAuth2SessionType `xorm:"index"`

	ClientID string        `xorm:"index"`
	Client   *OAuth2Client `xorm:"-"`

	RequestID         string
	RequestedAt       *time.Time
	GrantedAudience   []string
	RequestedAudience []string
	Form              string
	Data              []byte
	Subject           string
	Active            bool
}

// TableName returns the table's name
func (session OAuth2Session) TableName() string {
	return "oauth2_sessions"
}

// LoadClient will load the oauth2 client by the given client id
func (session *OAuth2Session) LoadClient() (err error) {
	session.Client, err = GetOAuth2ClientByClientID(session.ClientID)
	return
}

// GetOAuth2SessionBySignature returns the session with the given signature
func GetOAuth2SessionBySignature(signature string) (session *OAuth2Session, err error) {
	session = new(OAuth2Session)
	_, err = x.ID(calculateSignature(signature)).Get(&session)
	return
}

// CreateOAuth2Session creates the given session in the database
func CreateOAuth2Session(session *OAuth2Session) error {
	// calculate hash to prevent to big varchars
	session.Signature = calculateSignature(session.Signature)
	_, err := x.Insert(session)
	return err
}

// DeleteOAuth2SessionBySignature deletes the session with the given signature.
func DeleteOAuth2SessionBySignature(signature string) error {
	_, err := x.ID(calculateSignature(signature)).Delete(&OAuth2Session{})
	return err
}

// InvalidateOAuth2Session invalidates the session with the given signature
func InvalidateOAuth2Session(signature string) error {
	_, err := x.ID(calculateSignature(signature)).Cols("active").Update(&OAuth2Session{Active: false})
	return err
}
