// Vikunja is a to-do list application to facilitate your life.
// Copyright 2018-2020 Vikunja and contributors. All rights reserved.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package oauth

import (
	"code.vikunja.io/api/pkg/log"
	"code.vikunja.io/api/pkg/models"
	"encoding/json"
	"github.com/ory/fosite"
	"github.com/pkg/errors"
	"net/url"
)

func createSession(sessionType models.OAuth2SessionType, signature string, request fosite.Requester) error {
	session, err := requestToSession(sessionType, signature, request)
	if err != nil {
		return err
	}
	return models.CreateOAuth2Session(session)
}

func requestToSession(sessionType models.OAuth2SessionType, signature string, req fosite.Requester) (*models.OAuth2Session, error) {
	var subject string
	if req.GetSession() == nil {
		log.Debug("Empty session detected while converting requester to session")
	} else {
		subject = req.GetSession().GetSubject()
	}

	session, err := json.Marshal(req.GetSession())
	if err != nil {
		return nil, errors.WithStack(err)
	}
	requestedAt := req.GetRequestedAt()
	return &models.OAuth2Session{
		Signature:         signature,
		Type:              sessionType,
		ClientID:          req.GetClient().GetID(),
		Client:            nil,
		RequestID:         req.GetID(),
		RequestedAt:       &requestedAt,
		GrantedAudience:   req.GetGrantedAudience(),
		RequestedAudience: req.GetRequestedAudience(),
		Form:              req.GetRequestForm().Encode(),
		Data:              session,
		Subject:           subject,
		Active:            true,
	}, nil
}

func findSessionBySignature(signature string, fositeSession fosite.Session) (fosite.Requester, error) {
	sess, err := models.GetOAuth2SessionBySignature(signature)
	if err != nil {
		return nil, errors.WithStack(err)
	}
	if !sess.Active {
		if sess.Type == models.SessionTypeAuthorizationCode {
			return nil, errors.WithStack(fosite.ErrInvalidatedAuthorizeCode)
		}
		return nil, errors.WithStack(fosite.ErrInactiveToken)
	}
	return sessionToRequest(sess, fositeSession)
}

func sessionToRequest(sess *models.OAuth2Session, fositeSession fosite.Session) (fosite.Requester, error) {
	if sess.Data != nil {
		if err := json.Unmarshal(sess.Data, fositeSession); err != nil {
			return nil, errors.WithStack(err)
		}
	} else {
		log.Debug("empty session detected in sessionToRequest")
	}
	if sess.Client == nil || sess.ClientID != sess.Client.ClientID {
		if err := sess.LoadClient(); err != nil {
			return nil, errors.WithStack(err)
		}
	}
	form, err := url.ParseQuery(sess.Form)
	if err != nil {
		return nil, errors.WithStack(err)
	}

	return &fosite.Request{
		ID:          sess.RequestID,
		RequestedAt: *sess.RequestedAt,
		Client:      &fositeOAuth2Client{*sess.Client},
		// FIXME scopes are currently not supported
		RequestedScope:    make(fosite.Arguments, 0, 0),
		GrantedScope:      make(fosite.Arguments, 0, 0),
		Form:              form,
		Session:           fositeSession,
		RequestedAudience: sess.RequestedAudience,
		GrantedAudience:   sess.GrantedAudience,
	}, nil
}
