package oauth

import (
	"context"
	"github.com/pkg/errors"
	"time"

	"github.com/ory/fosite"

	"code.vikunja.io/api/pkg/models"
)

type Storage struct{}

func newStorage() *Storage {
	return &Storage{}
}

func (s *Storage) GetClient(ctx context.Context, id string) (fosite.Client, error) {
	client, err := models.GetOAuth2ClientByClientID(id)
	if err != nil {
		return nil, err
	}
	return &fositeOAuth2Client{OAuth2Client: *client}, nil
}

func (s *Storage) ClientAssertionJWTValid(ctx context.Context, jti string) error {
	if isBlocked, err := models.IsJTIBlocked(jti); err != nil {
		return err
	} else if isBlocked {
		return errors.WithStack(fosite.ErrJTIKnown)
	}
	return nil
}

func (s *Storage) SetClientAssertionJWT(ctx context.Context, jti string, exp time.Time) error {
	// Clean up the database
	if err := models.DeleteExpiredJTIs(); err != nil {
		return err
	}
	return models.BlockOAuth2JTI(jti, &exp)
}

func (s *Storage) CreateAuthorizeCodeSession(ctx context.Context, code string, request fosite.Requester) (err error) {
	return createSession(models.SessionTypeAuthorizationCode, code, request)
}

func (s *Storage) GetAuthorizeCodeSession(ctx context.Context, code string, session fosite.Session) (request fosite.Requester, err error) {
	return findSessionBySignature(code, session)
}

func (s *Storage) InvalidateAuthorizeCodeSession(ctx context.Context, code string) (err error) {
	return models.InvalidateOAuth2Session(code)
}

func (s *Storage) CreateAccessTokenSession(ctx context.Context, signature string, request fosite.Requester) (err error) {
	return createSession(models.SessionTypeAccessToken, signature, request)
}

func (s *Storage) GetAccessTokenSession(ctx context.Context, signature string, session fosite.Session) (request fosite.Requester, err error) {
	return findSessionBySignature(signature, session)
}

func (s *Storage) DeleteAccessTokenSession(ctx context.Context, signature string) (err error) {
	return models.DeleteOAuth2SessionBySignature(signature)
}

func (s *Storage) CreateRefreshTokenSession(ctx context.Context, signature string, request fosite.Requester) (err error) {
	return createSession(models.SessionTypeRefreshToken, signature, request)
}

func (s *Storage) GetRefreshTokenSession(ctx context.Context, signature string, session fosite.Session) (request fosite.Requester, err error) {
	return findSessionBySignature(signature, session)
}

func (s *Storage) DeleteRefreshTokenSession(ctx context.Context, signature string) (err error) {
	return models.DeleteOAuth2SessionBySignature(signature)
}
